/*
Intellicart Binary to ROM file conversion Program
Chad Schell - 6/19/2000

This code may be freely distributed and modified for use in the creation of new programs
for communication with the Intellicart.

Use of this code for other purposes in ok, but it's pretty clunky and you can probably find a 
better set of code to start from.
*/


#include <string.h>
#include <stdlib.h>
#include <stdio.h>

typedef struct ROM_SETUP
{
	unsigned int Number_Sections;
	unsigned short Boundaries[40];
	unsigned char Enable_Table[16];
	unsigned char Fine_Range[32];

} ROM_SETUP;


/*Function Declarations*/
void write_tables(FILE *outfile, ROM_SETUP *setup);
void write_crc(FILE *outfile, unsigned short crc);
void crc_init(unsigned short *reg);
void crc_calc(unsigned short *reg, unsigned char x);
void create_file_names(char *romname, char *cfgname, char *outname);
int Parse_Config_File(FILE *Cfg, ROM_SETUP *setup);
int Create_ROM_Send_File(char *romname, char *outname);
void ROM_Table_Update(ROM_SETUP *setup, unsigned short start_addr,
	unsigned short stop_addr, unsigned char type);

unsigned short crctable[256] =
{
 0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7,
 0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE, 0xF1EF,
 0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6,
 0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C, 0xF3FF, 0xE3DE,
 0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485,
 0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE, 0xF5CF, 0xC5AC, 0xD58D,
 0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4,
 0xB75B, 0xA77A, 0x9719, 0x8738, 0xF7DF, 0xE7FE, 0xD79D, 0xC7BC,
 0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823,
 0xC9CC, 0xD9ED, 0xE98E, 0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B,
 0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12,
 0xDBFD, 0xCBDC, 0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A,
 0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41,
 0xEDAE, 0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49,
 0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70,
 0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78,
 0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E, 0xE16F,
 0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067,
 0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C, 0xE37F, 0xF35E,
 0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256,
 0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E, 0xE54F, 0xD52C, 0xC50D,
 0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
 0xA7DB, 0xB7FA, 0x8799, 0x97B8, 0xE75F, 0xF77E, 0xC71D, 0xD73C,
 0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634,
 0xD94C, 0xC96D, 0xF90E, 0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB,
 0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3,
 0xCB7D, 0xDB5C, 0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A,
 0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92,
 0xFD2E, 0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9,
 0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
 0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8,
 0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1, 0x1EF0
};

/*Main's only purpose is to read the command line and call the rest of the code*/
int main(int argc, char *argv[])
{
	

	char romname[200],outname[200];

/*Must supply a binary file name, and nothing else.*/

	if(argc < 2) exit(1);
	if(argc > 2) exit(1);
	
/*Generate binary file name - add .bin extension*/
	strcpy(romname,argv[1]);
	strcat(romname,".bin");

/*Create the ROM file*/
	return Create_ROM_Send_File(romname,outname);
}


/*The bulk of the code.  Performs the actual conversion of the binary file to the ROM file*/
int Create_ROM_Send_File(char *romname, char *outname)
{
	/*Mapping represents that ROM sections were declared in a configuration file*/
	/*A zero means that they were not declared*/
	int mapping = 0;
	char cfgname[200];
	FILE *romfile, *configfile, *outfile;
	ROM_SETUP setup;
	unsigned int i;
	int j;
	unsigned char temp;
	int segment_size;
	long filesize;

	unsigned short crc;


	/*Initlialize Enable Tables*/
	for(i = 0; i < 16; i++)
		setup.Enable_Table[i] = 0;
	for(i = 0; i < 32; i++)
	{
		setup.Fine_Range[i] = 0x07;
	}

	/*Create the configuration and ROM file names*/
	create_file_names(romname, cfgname, outname);
	

	/*Open the binary file - abort if it doesn't exist*/
	if((romfile = fopen(romname,"rb"))== NULL)
	{
		return(1);
	}

	/*Open the ROM file for writing - overwrite any existing ROM file of the same name*/
	if((outfile = fopen(outname,"wb"))==NULL)
	{
		return(1);
	}

	/*This is necessary for auto baud rate detection*/
	fputc(0xA8,outfile);

	/*Is there a configuration file?*/
	if((configfile = fopen(cfgname,"r"))!=NULL)
	{
		/*Yes - parse it*/
		mapping = Parse_Config_File(configfile, &setup);
		fclose(configfile);
	}

	/*Did a configuration file with ROM segments exist?*/
	if(mapping != 0)
	{
		/*Yes - copy those segments from the binary file to the ROM file*/
		/*Write the number of ROM segments that will be in the ROM file*/
		fputc(setup.Number_Sections,outfile);
		/*Write it's one's complement for error detection*/
		fputc(~setup.Number_Sections,outfile);

		/*Write the ROM segments one a time*/
		for(i = 0; i < setup.Number_Sections; i++)
		{
			/*Initialize the CRC for each new ROM segment*/
			crc_init(&crc);

			/*Write the start and stop address of the ROM segment, and update the CRC*/
			temp = setup.Boundaries[i*2] / 0x100;
			crc_calc(&crc,temp);
			fputc(temp, outfile);
			temp = setup.Boundaries[i*2+1] / 0x100;
			crc_calc(&crc,temp);
			fputc(temp,outfile);

			/*Determine the size of the ROM segment, copy it to the ROM file, and update
			the CRC with each new byte of data*/
			segment_size = 2*( setup.Boundaries[i*2+1]-setup.Boundaries[i*2]+1);
			for(j = 0; j < segment_size; j++)
			{
				fread(&temp,1,1,romfile);
				fputc(temp, outfile);
				crc_calc(&crc,temp);
			}
			/*Write the CRC for this segment*/
			write_crc(outfile, crc);
		}
		/*Write the enable tables - these were created while parsing the configuration file*/
		write_tables(outfile, &setup);
	}

	else
	{
		/*No ROM segments were defined in the configuration file.  Attempt to fit the binary
		file to one of the standard cartridge layouts.  Update the Enable tables accordingly*/
		
		/*Determine file size*/
		fseek(romfile,0,SEEK_END);
		filesize = ftell(romfile);
		fseek(romfile,0,SEEK_SET);
		if(filesize == 0x2000)  /*4K Word ROM*/
		{
			fputc(1,outfile);
			fputc((unsigned char)(~1),outfile);
			crc_init(&crc);
			fputc(0x50,outfile);
			crc_calc(&crc,0x50);
			fputc(0x5F,outfile);
			crc_calc(&crc,0x5F);
			for(j = 0; j < 8192; j++)
			{
				fread(&temp,1,1,romfile);
				fputc(temp, outfile);
				crc_calc(&crc,temp);
			}
			write_crc(outfile, crc);
			/*4K Word carts occupy from 0x5000-0x5FFF*/
			ROM_Table_Update(&setup, 0x5000, 0x5FFF, 0x01);
			write_tables(outfile, &setup);
		}
		if(filesize == 0x4000) /*8K Word ROM*/
		{
			fputc(1,outfile);
			fputc((unsigned char)(~1),outfile);
			crc_init(&crc);
			fputc(0x50,outfile);
			crc_calc(&crc,0x50);
			fputc(0x6F,outfile);
			crc_calc(&crc,0x6F);
			for(j = 0; j < 16384; j++)
			{
				fread(&temp,1,1,romfile);
				fputc(temp, outfile);
				crc_calc(&crc,temp);
			}
			write_crc(outfile, crc);
			ROM_Table_Update(&setup, 0x5000, 0x6FFF, 0x01);
			write_tables(outfile, &setup);
		}
		if(filesize == 0x6000) /*12K Word ROM*/
		{
			fputc(2,outfile);
			fputc((unsigned char)(~2),outfile);
			crc_init(&crc);
			fputc(0x50,outfile);
			crc_calc(&crc,0x50);
			fputc(0x6F,outfile);
			crc_calc(&crc,0x6F);
			for(j = 0; j < 16384; j++)
			{
				fread(&temp,1,1,romfile);
				fputc(temp, outfile);
				crc_calc(&crc,temp);
			}
			write_crc(outfile, crc);
			ROM_Table_Update(&setup, 0x5000, 0x6FFF, 0x01);
			crc_init(&crc);
			fputc(0xD0,outfile);
			crc_calc(&crc,0xD0);
			fputc(0xDF,outfile);
			crc_calc(&crc,0xDF);
			for(j = 0; j < 8192; j++)
			{
				fread(&temp,1,1,romfile);
				fputc(temp, outfile);
				crc_calc(&crc,temp);
			}
			write_crc(outfile, crc);
			ROM_Table_Update(&setup, 0xD000, 0xDFFF, 0x01);
			write_tables(outfile, &setup);
		}
		if(filesize == 0x8000) /*16K Word ROM*/
		{
			fputc(3,outfile);
			fputc((unsigned char)(~3),outfile);
			crc_init(&crc);
			fputc(0x50,outfile);
			crc_calc(&crc,0x50);
			fputc(0x6F,outfile);
			crc_calc(&crc,0x6F);
			for(j = 0; j < 16384; j++)
			{
				fread(&temp,1,1,romfile);
				fputc(temp, outfile);
				crc_calc(&crc,temp);
			}
			write_crc(outfile, crc);
			ROM_Table_Update(&setup, 0x5000, 0x6FFF, 0x01);
			
			crc_init(&crc);
			fputc(0xD0,outfile);
			crc_calc(&crc,0xD0);
			fputc(0xDF,outfile);
			crc_calc(&crc,0xDF);
			for(j = 0; j < 8192; j++)
			{
				fread(&temp,1,1,romfile);
				fputc(temp, outfile);
				crc_calc(&crc,temp);
			}
			write_crc(outfile, crc);
			ROM_Table_Update(&setup, 0xD000, 0xDFFF, 0x01);

			crc_init(&crc);
			fputc(0xF0,outfile);
			crc_calc(&crc,0xF0);
			fputc(0xFF,outfile);
			crc_calc(&crc,0xFF);
			for(j = 0; j < 8192; j++)
			{
				fread(&temp,1,1,romfile);
				fputc(temp, outfile);
				crc_calc(&crc,temp);
			}
			write_crc(outfile, crc);
			ROM_Table_Update(&setup, 0xF000, 0xFFFF, 0x01);
			write_tables(outfile, &setup);
		}

		/*No ROM segments were defined in a configuration file, so if it was not a standard
		file size, there is an error with the binary file*/
		if(( (filesize == 0x2000) || (filesize == 0x4000) || (filesize == 0x6000) 
			|| (filesize == 0x8000) ) == 0)
		{
			printf("Bad File size\n");fflush(stdout);
			fclose(romfile);
			fclose(outfile);
			return(1);
		}
	}
	/*Close the files*/
	fclose(romfile);
	fclose(outfile);

	/*Done*/
	return 0;
}

/*This routine writes the CRC value to the ROM file in the proper format*/
void write_crc(FILE *outfile, unsigned short crc)
{
	int temp;
	temp = (crc & 0xFF00) >> 8;
	fputc(temp, outfile);
	temp = (crc & 0xFF);
	fputc(temp,outfile);
	return;
}

/*This routine writes the Enable tables to the ROM file in the proper format.
It also calculates their CRC and writes that to the ROM file as well.*/
void write_tables(FILE *outfile, ROM_SETUP *setup)
{
	unsigned short crc;
	int j;
	crc_init(&crc);
	for(j = 0; j < 16; j ++)
	{
		crc_calc(&crc,setup->Enable_Table[j]);
	}
	fwrite(setup->Enable_Table,1,16,outfile);
	for(j = 0; j < 32; j ++)
	{
		crc_calc(&crc,setup->Fine_Range[j]);
	}
	fwrite(setup->Fine_Range,1,32,outfile);
	write_crc(outfile, crc);
}

/*A VERY simple parser to read configuration files - NOT robust at all.
Spacing in commands must EXACTLY match the patterns specified*/
int Parse_Config_File(FILE *Cfg, ROM_SETUP *setup)
{
//	int i,j,min;
	char line[200];
	unsigned rom1,rom2,addr;
	int mapping = 0;
	setup->Number_Sections = 0;

	while(fgets(line,200,Cfg)!=NULL)
	{
		/*[mapping] -- defines ROM segments, sets read access to those segments
		Addresses must follow in the format $xxxx - $yyyy = $zzzz
		No comment only lines can appear between [mapping] and the commands*/

		if(!strncmp(line,"[mapping]",9)) /* mapping line detected*/
		{
			mapping = 1;
			while(fgets(line,200,Cfg)!=NULL)
			{
				if(line[0] == '$') /*Address command line*/
				{	
					sscanf(line,"$%x - $%x = $%x",&rom1,&rom2,&addr);
					setup->Boundaries[(setup->Number_Sections)*2] = addr;
					setup->Boundaries[(setup->Number_Sections)*2+1] = (unsigned short)(rom2-rom1+addr);
					/*Modify enable tables for read access*/
					ROM_Table_Update(setup,
						setup->Boundaries[(setup->Number_Sections)*2],
						setup->Boundaries[(setup->Number_Sections)*2+1],0x01);
					setup->Number_Sections++;
				}
				else
					break;
			}
		}
		/*[Preload] -- same as [mapping] but does not set Enable tables for read access*/
		if(!strncmp(line,"[preload]",9)) /* preload line detected*/
		{
			mapping = 1;
			while(fgets(line,200,Cfg)!=NULL)
			{
				if(line[0] == '$') /*Address command line*/
				{	
					sscanf(line,"$%x - $%x = $%x",&rom1,&rom2,&addr);
					setup->Boundaries[(setup->Number_Sections)*2] = addr;
					setup->Boundaries[(setup->Number_Sections)*2+1] = (unsigned short)(rom2-rom1+addr);
					setup->Number_Sections++;
				}
				else
					break;
			}
		}
		/*[memattr] -- used to define RAM areas.  Sets read and write access for these areas
		Addresses must follow in the format $xxxx -$yyyy = $zzzz
		Note the lack of a space after the hyphon.  This matches the existing .cfg files,
		and this is an example of how unrobust this parser is*/
		if(!strncmp(line,"[memattr]",9)) /* memattr line detected*/
		{
			while(fgets(line,200,Cfg)!=NULL)
			{
				if(line[0] == '$') /*Address command line*/
				{	
					sscanf(line,"$%x -$%x = RAM %d",&rom1,&rom2,&addr);
					/*Modify Tables to enable RAM in this region*/
					/*There is no provision for 8 vs. 16 bit RAM.*/
					ROM_Table_Update(setup, (unsigned short)rom1, (unsigned short)rom2, 0x03);
				}
				else
					break;
			}
		}
		/*[bankswitch] -- used to enable bankswitching on a given window, 
		also enables read access to that window.
		Addresses must follow in the format $xxxx - $yyyy*/
		if(!strncmp(line,"[bankswitch]",9)) /* bankswitch line detected*/
		{
			while(fgets(line,200,Cfg)!=NULL)
			{
				if(line[0] == '$') /*ddress command line */
				{	
					sscanf(line,"$%x - $%x",&rom1,&rom2);
					/*Load Bankswitch Table with starting address and modify enable tables*/
					ROM_Table_Update(setup, (unsigned short)rom1, (unsigned short)rom2, 0x09);
				}
				else
					break;
			}
		}


	}

	if(mapping == 0)
		return 0;
	else
		return 1;
}

//This code handles the address convolution to place values in the Enable tables.
void ROM_Table_Update(ROM_SETUP *setup, unsigned short start_addr,
	unsigned short stop_addr, unsigned char enable_bits)
{
	unsigned short addr;
	unsigned char high_addr, index;
	addr = start_addr;
	while(addr < stop_addr)
	{
		high_addr = (addr & 0xFF00) >> 8;
		index = (high_addr & 0xF0) >> 4;
		if((high_addr & 0x08) == 0)
		{
			setup->Enable_Table[index] |= enable_bits;
		}
		else
		{
			setup->Enable_Table[index] |= (enable_bits << 4);
		}
		addr += 0x0800;
		if(addr == 0) break;
	}
	/*Do Fine Range Restrictions if Necessary*/
	/*Start Address*/
	addr = (start_addr & 0x0700) >> 8;
	if(addr != 0)
	{
		high_addr = (start_addr & 0xFF00) >> 8;
		index = (high_addr & 0xF0) >> 4;
		if((high_addr & 0x08) == 0)
		{
			setup->Fine_Range[index] |= (addr << 4);
		}
		else
		{
			setup->Fine_Range[index+16] |= (addr << 4);
		}
	}
	/*Stop Address*/
	addr = (stop_addr & 0x0700) >> 8;
	if(addr != 7)
	{
		high_addr = (stop_addr & 0xFF00) >> 8;
		index = (high_addr & 0xF0) >> 4;
		if((high_addr & 0x08) == 0)
		{
			setup->Fine_Range[index] &= 0xF0;
			setup->Fine_Range[index] |= addr;
		}
		else
		{
			setup->Fine_Range[index+16] &= 0xF0;
			setup->Fine_Range[index+16] |= addr;
		}
	}

}

/*Creates the .rom and .cfg filenames from the .bin filename*/
void create_file_names(char *romname, char *cfgname, char *outname)
{
	int namesize = strlen(romname);
	strncpy(cfgname,romname,namesize-3);
	strncpy(outname,romname,namesize-3);
	cfgname[namesize-3] = '\0';
	outname[namesize-3] = '\0';
	strcat(cfgname,"cfg");
	strcat(outname,"rom");
}

/*Initializes the CRC to it's starting value*/
void crc_init(unsigned short *reg)
{
	*reg = 0xFFFF;
}

/*Computes the CRC, using the CRC table at the top of this code*/
void crc_calc(unsigned short *reg, unsigned char x)
{
	*reg = (*reg<<8) ^ crctable[(*reg >> 8) ^ x];
}